# -*- coding: utf-8 -*-


class DTShellConstants(object):
    COMMANDS_REPO_OWNER = 'duckietown'
    COMMANDS_REPO_NAME = 'duckietown-shell-commands'
    COMMANDS_REPO_BRANCH = 'master'
    COMMANDS_REMOTE_URL = 'https://github.com/%s/%s' % (COMMANDS_REPO_OWNER, COMMANDS_REPO_NAME)
    ROOT = '~/.dt-shell/'
    ENV_COMMANDS = 'DTSHELL_COMMANDS'

    DT1_TOKEN_CONFIG_KEY = 'token_dt1'
    # DT1_TOKEN_ENV_VARIABLE = 'DUCKIETOWN_TOKEN'

    CONFIG_DOCKER_USERNAME = 'docker_username'



# coding=utf-8
import json
import os
import time

import dateutil.parser
import six
import termcolor

from . import dtslogger
from .utils import indent


class Storage(object):
    done = False


DEFAULT_DTSERVER = 'https://challenges.duckietown.org/v4'


def get_duckietown_server_url():
    V = 'DTSERVER'

    if V in os.environ:
        use = os.environ[V]
        if not Storage.done:
            if use != DEFAULT_DTSERVER:
                msg = 'Using server %s instead of default %s' % (use, DEFAULT_DTSERVER)
                dtslogger.info(msg)
            Storage.done = True
        return use
    else:
        return DEFAULT_DTSERVER


class RequestException(Exception):
    pass


class ConnectionError(RequestException):
    """ The server could not be reached or completed request or
        provided an invalid or not well-formatted answer. """


class RequestFailed(RequestException):
    """
        The server said the request was invalid.

        Answered  {'ok': False, 'error': msg}
    """


DEFAULT_TIMEOUT = 5


def make_server_request(token, endpoint, data=None, method='GET', timeout=DEFAULT_TIMEOUT,
                        suppress_user_msg=False):
    """
        Raise RequestFailed or ConnectionError.

        Returns the result in 'result'.
    """

    from six.moves import urllib
    # import urllib.request

    server = get_duckietown_server_url()
    url = server + endpoint

    headers = {}
    if token is not None:
        headers['X-Messaging-Token'] = token

    if data is not None:
        data = json.dumps(data)
        if six.PY3:
            data = data.encode('utf-8')
    t0 = time.time()
    # dtslogger.info('server request with timeout %s' % timeout)
    req = urllib.request.Request(url, headers=headers, data=data)
    req.get_method = lambda: method

    try:
        # dtslogger.info('urlopen')
        res = urllib.request.urlopen(req, timeout=timeout)
        # dtslogger.info('read')
        data_read = res.read()
    except urllib.error.HTTPError as e:
        msg = 'Operation failed for %s' % url
        err_msg = e.read().decode("utf-8")
        msg += f'\n\n{err_msg}'
        raise ConnectionError(msg)
    except urllib.error.URLError as e:
        msg = 'Cannot connect to server %s' % url
        raise ConnectionError(msg) from e

    delta = time.time() - t0
    # dtslogger.info('server request took %.1f seconds' % delta)

    data_s = data_read.decode('utf-8')
    try:
        result = json.loads(data_s)
    except ValueError as e:
        msg = 'Cannot read answer from server.'
        msg += '\n\n' + indent(data_s, '  > ')
        raise ConnectionError(msg) from e

    if not isinstance(result, dict) or 'ok' not in result:
        msg = 'Server provided invalid JSON response. Expected a dict with "ok" in it.'
        msg += '\n\n' + indent(data, '  > ')
        raise ConnectionError(msg)

    if 'user_msg' in result and not suppress_user_msg:
        user_msg = result['user_msg']
        if user_msg:
            s = []
            lines = user_msg.strip().split('\n')
            prefix = u'message from server: '
            p2 = u': '.rjust(len(prefix))
            print('')

            for i, l in enumerate(lines):
                p = prefix if i == 0 else p2
                # l = termcolor.colored(l, 'blue')
                s.append(termcolor.colored(p, attrs=['dark']) + l)
            from dt_shell.cli import dts_print

            dts_print('\n'.join(s))

    if result['ok']:
        if 'result' not in result:
            msg = 'Server provided invalid JSON response. Expected a field "result".'
            msg += '\n\n' + indent(result, '  > ')
            raise ConnectionError(msg)
        return result['result']
    else:
        msg = result.get('msg', 'no error message in %s ' % result)
        msg = 'Failed request for %s:\n%s' % (url, msg)
        raise RequestFailed(msg)


def get_dtserver_user_info(token):
    """ Returns a dictionary with information about the user """
    endpoint = '/info'
    method = 'GET'
    data = None
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_submit(token, queue, data):
    endpoint = '/submissions'
    method = 'POST'
    data = {'queue': queue, 'parameters': data}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_retire(token, submission_id):
    endpoint = '/submissions'
    method = 'DELETE'
    data = {'submission_id': submission_id}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_get_user_submissions(token):
    """ Returns a dictionary with information about the user submissions """
    endpoint = '/submissions'
    method = 'GET'
    data = {}
    submissions = make_server_request(token, endpoint, data=data, method=method)

    for v in submissions.values():
        for k in ['date_submitted', 'last_status_change']:
            v[k] = dateutil.parser.parse(v[k])
    return submissions



import getpass
import grp
import os
import subprocess
import sys

from whichcraft import which

from dt_shell import dtslogger
from .constants import DTShellConstants
from .exceptions import InvalidEnvironment, UserError


def running_with_sudo():
    if 'SUDO_USER' in os.environ:
        return True
    return False


def abort_if_running_with_sudo():
    if running_with_sudo():
        msg = '''\
Do not run dts using "sudo".'

As a matter of fact, do not run anything with "sudo" unless instructed to do so.\
'''
        raise UserError(msg)


def check_docker_environment():
    username = getpass.getuser()
    from . import dtslogger
    dtslogger.debug('Checking docker environment for user %s' % username)

    check_executable_exists('docker')

    check_user_in_docker_group()
    #
    # if on_linux():
    #
    #     if username != 'root':
    #         check_user_in_docker_group()
    #     # print('checked groups')
    # else:
    #     dtslogger.debug('skipping env check because not on Linux')

    try:
        import docker
    except Exception as e:
        msg = 'Could not import package docker:\n%s' % e
        msg += '\n\nYou need to install the package'
        raise InvalidEnvironment(msg)

    if 'DOCKER_HOST' in os.environ:
        msg = 'Note that the variable DOCKER_HOST is set to "%s"' % os.environ['DOCKER_HOST']
        dtslogger.warning(msg)

    try:
        client = docker.from_env()

        containers = client.containers.list(filters=dict(status='running'))

        # dtslogger.debug(json.dumps(client.info(), indent=4))

    except Exception as e:
        msg = 'I cannot communicate with Docker:\n%s' % e
        msg += '\n\nMake sure the docker service is running.'
        raise InvalidEnvironment(msg)

    return client


def on_linux():
    return sys.platform.startswith('linux')


def check_executable_exists(cmdname):
    p = which(cmdname)
    if p is None:
        msg = 'Could not find executable "%s".' % cmdname
        raise InvalidEnvironment(msg)


def check_user_in_docker_group():
    # first, let's see if there exists a group "docker"
    group_names = [g.gr_name for g in grp.getgrall()]
    G = 'docker'
    if G not in group_names:
        msg = 'No group %s defined.' % G
        dtslogger.warning(msg)
    else:
        group_id = grp.getgrnam(G).gr_gid
        my_groups = os.getgroups()
        if group_id not in my_groups:
            msg = 'My groups are %s and "%s" group is %s ' % (my_groups, G, group_id)
            msg += '\n\nNote that when you add a user to a group, you need to login in and out.'
            dtslogger.warning(msg)
    #
    #     active_groups = get_active_groups(username=None)
    #
    # if name not in active_groups:
    #     msg = 'The user is not in group "%s".' % name
    #     msg += '\n\nIt belongs to groups: %s.' % u", ".join(sorted(active_groups))
    #
    #     msg += '\n\nNote that when you add a user to a group, you need to login in and out.'
    #
    #     if True:
    #         dtslogger.warning(msg)
    #     else:
    #         raise InvalidEnvironment(msg)


def get_active_groups(username=None):
    cmd = ['groups']

    if username:
        cmd.append(username)

    try:
        stdout = subprocess.check_output(['groups'])
        # res = system_cmd_result('.', cmd,
        #                         display_stdout=False,
        #                         display_stderr=False,
        #                         raise_on_error=True,
        #                         capture_keyboard_interrupt=False,
        #                         env=None)
    except subprocess.CalledProcessError as e:
        raise Exception(str(e))
    active_groups = stdout.decode().split()

    return active_groups


def get_dockerhub_username(shell=None):
    if shell is None:
        from .cli import DTShell
        shell = DTShell()
    k = DTShellConstants.CONFIG_DOCKER_USERNAME
    if k not in shell.config:
        msg = 'Please set docker username using\n\n dts challenges config --docker-username <USERNAME>'
        raise InvalidEnvironment(msg)

    username = shell.config[k]
    return username



# -*- coding: utf-8 -*-

import logging

import yaml

from dt_shell.utils import format_exception
from .exceptions import *

logging.basicConfig()

dtslogger = logging.getLogger('dts')
dtslogger.setLevel(logging.DEBUG)

__version__ = '4.0.5'


dtslogger.info('duckietown-shell %s' % __version__)

import sys

if sys.version_info < (3, 6):
    msg = 'duckietown-shell works with Python 3.6 and later. Detected %s.' % str(sys.version_info)
    sys.exit(msg)

class OtherVersions(object):
    name2versions = {}



import termcolor

from .cli import DTShell, dts_print

from .dt_command_abs import DTCommandAbs
from .dt_command_placeholder import DTCommandPlaceholder


def cli_main():
    from .col_logging import setup_logging_color
    setup_logging_color()

    known_exceptions = (InvalidEnvironment, CommandsLoadingException)
    try:
        cli_main_()
    except UserError as e:
        msg = str(e)
        dts_print(msg, 'red')
        print_version_info()
        sys.exit(1)
    except known_exceptions as e:
        msg = str(e)
        dts_print(msg, 'red')
        print_version_info()
        sys.exit(1)
    except SystemExit:
        raise
    except KeyboardInterrupt:
        dts_print('User aborted operation.')
        pass
    except BaseException as e:
        msg = format_exception(e)
        dts_print(msg, 'red', attrs=['bold'])
        print_version_info()
        sys.exit(2)


def print_version_info():
    v = OtherVersions.name2versions
    v['python'] = sys.version
    v['duckietown-shell'] = __version__

    msg = '''\
Please report that you are using:

%s
''' % yaml.dump(v, default_flow_style=False)
    dts_print(msg, 'red', attrs=['dark'])


def cli_main_():
    from .env_checks import abort_if_running_with_sudo
    abort_if_running_with_sudo()
    # Problems with a step in the Duckiebot operation manual?
    #
    #     Report here: https://github.com/duckietown/docs-opmanual_duckiebot/issues

    # TODO: register handler for Ctrl-C
    url = href("https://github.com/duckietown/duckietown-shell-commands/issues")
    msg = """

Problems with a command?

Report here: {url}

Troubleshooting:

- If some commands update fail, delete ~/.dt-shell/commands

- To reset the shell to "factory settings", delete ~/.dt-shell

  (Note: you will have to re-configure.)

    """.format(url=url)
    dts_print(msg)

    from .exceptions import InvalidEnvironment, UserError

    shell = DTShell()
    arguments = sys.argv[1:]

    if arguments:
        from dt_shell.utils import replace_spaces
        arguments = map(replace_spaces, arguments)
        cmdline = " ".join(arguments)
        shell.onecmd(cmdline)
    else:
        shell.cmdloop()


def href(x):
    return termcolor.colored(x, 'blue', attrs=['underline'])



import traceback

import six


def indent(s, prefix, first=None):
    s = str(s)
    assert isinstance(prefix, str)
    lines = s.split('\n')
    if not lines:
        return ''

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = ' ' * (m - len(prefix)) + prefix
    first = ' ' * (m - len(first)) + first

    # different first prefix
    res = ['%s%s' % (prefix, line.rstrip()) for line in lines]
    res[0] = '%s%s' % (first, lines[0].rstrip())
    return '\n'.join(res)


def raise_wrapped(etype, e, msg, compact=False, exc=None, **kwargs):
    """ Raises an exception of type etype by wrapping
        another exception "e" with its backtrace and adding
        the objects in kwargs as formatted by format_obs.

        if compact = False, write the whole traceback, otherwise just str(e).

        exc = output of sys.exc_info()
    """

    e = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)

    #     if exc is not None:
    #         _, _, trace = exc
    #         raise etype, e.args, trace
    #     else:
    raise e


def raise_wrapped_make(etype, e, msg, compact=False, **kwargs):
    """ Constructs the exception to be thrown by raise_wrapped() """
    assert isinstance(e, BaseException), type(e)
    assert isinstance(msg, str), type(msg)
    s = msg

    import sys
    if sys.version_info[0] >= 3:
        es = str(e)
    else:
        if compact:
            es = str(e)
        else:
            es = format_exception(e)

    s += '\n' + indent(es.strip(), '| ')

    return etype(s)


def check_isinstance(ob, expected, **kwargs):
    if not isinstance(ob, expected):
        kwargs['object'] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs):
    """ Raises an exception concerning ob having the wrong type. """
    e = 'Object not of expected type:'
    e += '\n  expected: %s' % str(expected)
    e += '\n  obtained: %s' % str(type(ob))
    # e += '\n' + indent(format_obs(kwargs), ' ')
    raise ValueError(e)


SPACE_TAG = 'SPACE_TAG'


def replace_spaces(x):
    return x.replace(' ', SPACE_TAG)


def undo_replace_spaces(x):
    return x.replace(SPACE_TAG, ' ')


def format_exception(e):
    if six.PY2:
        return traceback.format_exc(e)
    else:
        return traceback.format_exc()  # None, e)



import datetime
import json
import sys

import dateutil.parser
from future import builtins

from .duckietown_tokens import DuckietownToken, get_verify_key


def verify_a_token_main(args=None):
    try:
        if args is None:
            args = sys.argv[1:]

        if args:
            token_s = args[0]
        else:
            msg = 'Please enter token:\n> '
            token_s = builtins.input(msg)

        sys.stderr.write('Verifying token %r\n' % token_s)

        try:
            token = DuckietownToken.from_string(token_s)
        except ValueError:
            msg = "Invalid token format."
            sys.stderr.write(msg + '\n')
            sys.exit(3)

        vk = get_verify_key()
        ok = vk.verify(token.signature, token.payload)
        if not ok:
            msg = 'This is an invalid token; signature check failed.'
            sys.stderr.write(msg + '\n')
            sys.exit(5)

        try:
            data = json.loads(token.payload)
        except ValueError:
            msg = 'Invalid token format; cannot interpret payload %r.' % token.payload
            sys.stderr.write(msg + '\n')
            sys.exit(4)

        if not 'uid' in data or not 'exp' in data:
            msg = 'Invalid token format; missing fields from %s.' % data
            sys.stderr.write(msg + '\n')
            sys.exit(6)

        if data['uid'] == -1:
            msg = 'This is the sample token. Use your own token.'
            sys.stderr.write(msg + '\n')
            sys.exit(7)

        exp_date = dateutil.parser.parse(data['exp'])
        now = datetime.datetime.today()

        if exp_date < now:
            msg = 'This token has expired on %s' % exp_date
            sys.stderr.write(msg + '\n')
            sys.exit(6)

        o = dict()
        o['uid'] = data['uid']
        o['expiration'] = data['exp']
        msg = json.dumps(o)
        print(msg)
        sys.exit(0)

    except Exception as e:
        sys.stderr.write(str(e) + '\n')
        sys.exit(3)
